
#Global variables
#cache of the current queue
declare -a _lander_cached_queue=()
#current queue length
_lander_qsize=0
#disambiguator: different hosts/processes
_lander_queue_salt="$(hostname -s).$$"

# Set _lander_qsize to the number of files waiting in the queue
# Caches the current queue state.
lander_queue_size() {
    if [ "${#_lander_cached_queue[*]}" == 0 ]; then
	lander_queue_size_$LANDER_QUEUE_TYPE $@
    fi
    _lander_qsize="${#_lander_cached_queue[*]}"

    [ "$_lander_qsize" == 0 ] && return 1
    return 0
}
lander_queue_size_hdfs() {
    local landername=$1
    local files=$( $HDFS -lsr $LANDER_QUEUE_RAW_HDFS/$landername | 
	            awk '{print $8}' |  egrep '\/[0-9]{8}-[0-9]{6}-[0-9]+$' )
    [ -z "$files" ] && { _lander_cached_queue=(); _lander_qsize=0; return 1; }
    # sort files by name (but ignore the directories)
    local fnkey=$(echo "/x$files" 2>/dev/null | head -1 | tr '/' ' ' | wc -w) #should be num of '/' + 1
    _lander_cached_queue=( $(echo "$files" | sort -t '/' -k $fnkey) )
    _lander_qsize=${#_lander_cached_queue[*]}
    return 0
}
lander_queue_size_fs() {
    local landername=$1
    local files=$(find $LANDER_QUEUE_RAW_FS/$landername -regextype posix-extended \
	-regex '.*\/[0-9]{8}-[0-9]{6}-[0-9]+$' -print 2>/dev/null)
    [ -z "$files" ] && { _lander_cached_queue=(); _lander_qsize=0; return 1; }
    # sort files by name (but ignore the directories)
    local fnkey=$(echo "/x$files" 2>/dev/null | \
	head -1 | tr '/' ' ' | wc -w) #should be num of '/' + 1
    _lander_cached_queue=( $(echo "$files" | sort -t '/' -k $fnkey) )
    _lander_qsize=${#_lander_cached_queue[*]}
    return 0
}

#clear cached state of the queue
lander_queue_flush_cache() {
    _lander_cached_queue=()
    _lander_qsize=0
}

# lander_queue_file_older <filename> <ageseconds>
# Check if the file's modtime is at least ageseconds in the past
lander_queue_file_older() {
    lander_queue_file_older_$LANDER_QUEUE_TYPE $@
}
lander_queue_file_older_fs() {
    local filename=$1 ageseconds=$2 modtime
    modtime=$(stat -c%Y $filename) || return 1
    local now=$(date +%s)
    (( now - modtime > ageseconds ))
}
lander_queue_file_older_hdfs() {
    local filename=$1 ageseconds=$2 modtimems
    modtimems=$($HDFS -stat %Y $filename) || return 1
    local now=$(date +%s)
    (( now - modtimems/1000 > ageseconds ))
}

# Pop one file from the head of the queue, cache the current queue state.
# Sets 'reserved_local' to the file name on the local FS.
# Returns true if success or false if the queue is empty. 
#
# we want to be extremely careful not to read/write to shared fs if it
# can be helped.  So all our processing components read/write to /tmp
# and then copy results to shared fs, while retaining the result
# locally.
lander_queue_reserve_top() {
    local localtmp=$1
    lander_queue_reserve_top_$LANDER_QUEUE_TYPE $localtmp
}
lander_queue_reserve_top_hdfs() {
    local localtmp=$1
    while true; do
	#use queue_size not queue_size_hdfs, for caching
	lander_queue_size || \
	    { reserved_local=""; return 1; }
        local firstinq=${_lander_cached_queue[0]}
	[ -z "$firstinq" ] && { reserved_local=""; return 1; }
        _lander_cached_queue=( "${_lander_cached_queue[@]:1}" )
	_lander_qsize="${#_lander_cached_queue[*]}"
        local file=$(basename $firstinq)
        local lander=$(basename $(dirname $firstinq))
	local reservation="$LANDER_QUEUE_RESERVED_HDFS/$lander/$file-$_lander_queue_salt"
	#check if next seq number has been written
	local seqnum=${file##*-}
	local nseqlen=-${#seqnum}
	local nextseqnum="000000000000$(( 10#$seqnum+1 ))"
	nextseqnum=${nextseqnum:$nseqlen}
	if ! $HDFS -ls $LANDER_QUEUE_RAW_HDFS/$lander/*-$nextseqnum >&/dev/null; then
	    while true; do
		local now_s=$(date +%s)
		local modtime_ms=$($HDFS -stat %Y $firstinq)
		if ((now_s*1000 - modtime_ms >= LANDER_MINFILEAGE*1000)); then
	            #file at the head hasn't been modified for a while
		    break;
		else
		    sleep 60s
		fi
	    done
	    if ! $HDFS -ls $LANDER_QUEUE_RAW_HDFS/$lander/*-$nextseqnum >&/dev/null; then
		echo "WARN: working on $firstinq, while $nextseqnum hasn't been delivered" >&2
	    fi
	fi
        $HDFS -mv $firstinq $reservation 2>/dev/null \
            || { sleep 10; continue; }
	$HDFS -test -e $reservation 2>/dev/null \
	    || { sleep 10; continue; }
        mkdir -p $localtmp/$lander 2>/dev/null
        reserved_local="$localtmp/$lander/$file"
        $HDFS -get $reservation $reserved_local || {
	    echo "ERR: error hdfs -get $reservation $reserved_local" >&2
	    continue
	}
        return 0
    done
}
lander_queue_reserve_top_fs() {
    local localtmp=$1
    while true; do
	#use queue_size not queue_size_fs, for caching
	lander_queue_size || { reserved_local=""; return 1; }
        local firstinq=${_lander_cached_queue[0]}
        _lander_cached_queue=( "${_lander_cached_queue[@]:1}" )

        local file=$(basename $firstinq)
        local lander=$(basename $(dirname $firstinq))
	local reservation="$LANDER_QUEUE_RESERVED_FS/$lander/$file-$_lander_queue_salt"
	#check if next seq number has been written
	local seqnum=${file##*-}
	local nseqlen=-${#seqnum}
	local nextseqnum="000000000000$(( 10#$seqnum+1 ))"
	nextseqnum=${nextseqnum:$nseqlen}
	if ! [ -e $LANDER_QUEUE_RAW_FS/$lander/*-$nextseqnum 2>/dev/null ]; then
	    while true; do
		local now=$(date +%s)
		local modtime=$(/usr/bin/stat --format %Y $firstinq 2>/dev/null)
		if ((now - modtime >= $LANDER_MINFILEAGE)); then
	            #file at the head hasn't been modified for a while
		    break;
		else
		    sleep 60s
		fi
	    done
	    if ! [ -e $LANDER_QUEUE_RAW_FS/$lander/*-$nextseqnum 2>/dev/null ]; then
		echo "WARN: working on $firstinq, while $nextseqnum hasn't been delivered" >&2
	    fi
	fi
        mv $firstinq $reservation 2>/dev/null \
            || continue
	# sometimes mv exit status is zero, but move in fact
	#+fails (probably because of NFS attribute caching), 
	#+so we have to check for existence of the target.
	[ -e "$reservation" ] \
	    || continue
	mkdir -p $localtmp/$lander 2>/dev/null
	reserved_local="$localtmp/$lander/$file"
	ln $reservation $reserved_local 2>/dev/null \
	    || cp -p $reservation $reserved_local \
	    || { echo "Error copying $reservation to $reserved_local"; continue; }
        return 0
    done
}


# Put previously reserved with reserve_top_hdfs() file back into the queue.
# This can be useful e.g. if the file is to be processed by a different agent.
lander_queue_unreserve() {
    local file=$1
    lander_queue_unreserve_$LANDER_QUEUE_TYPE $file
}
lander_queue_unreserve_hdfs() {
    local file=$1
    local lander=$(basename $(dirname $file))
    local f=$(basename $file)
    local reservation="$LANDER_QUEUE_RESERVED_HDFS/$lander/$f-$_lander_queue_salt"

    $HDFS -mv $reservation $LANDER_QUEUE_RAW_HDFS/$lander || {
	echo "Cannot move $reservation back to $LANDER_QUEUE_RAW_HDFS/$lander" >&2
	return 1
    }
    rm -f $file
    lander_queue_flush_cache
    return 0
}
lander_queue_unreserve_fs() {
    local file=$1
    local lander=$(basename $(dirname $file))
    local f=$(basename $file)
    local reservation="$LANDER_QUEUE_RESERVED_FS/$lander/$f-$_lander_queue_salt"
    mv $reservation $LANDER_QUEUE_RAW_FS/$lander || {
	echo "Cannot move $reservation back to $LANDER_QUEUE_RAW_FS/$lander" >&2
	return 1
    }
    lander_queue_flush_cache
    rm -f $file
    return 0
}

lander_queue_wipe_source() {
    local lander=$1
    lander_queue_wipe_source_$LANDER_QUEUE_TYPE $lander
}

lander_queue_wipe_source_hdfs() {
    local lander=$1
    $HDFS -rm "$LANDER_QUEUE_RAW_HDFS/$lander/[0-9]*" >& /dev/null
}

lander_queue_wipe_source_fs() {
    local lander=$1
    #rm -f $LANDER_QUEUE_RAW_FS/$lander/[0-9]* >& /dev/null
    find $LANDER_QUEUE_RAW_FS/$lander/ -regextype posix-extended \
	-regex '.*\/[0-9]{8}-[0-9]{6}-[0-9]+$' -exec rm -f '{}' \;
}

# The file has been entirely processed, remove it from HDFS, as well as from
# local tmp.
lander_queue_release() {
    local file=$1
    lander_queue_release_$LANDER_QUEUE_TYPE $file
    /bin/rm -f $file
}
lander_queue_release_hdfs() {
    local file=$1
    local lander=$(basename $(dirname $file))
    local f=$(basename $file)
    local reservation="$LANDER_QUEUE_RESERVED_HDFS/$lander/$f-$_lander_queue_salt"
    $HDFS -rm $reservation >& /dev/null || {
        echo "Could not remove reservation $reservation" >&2
	return 1
    }
    return 0
}
lander_queue_release_fs() {
    local file=$1
    local lander=$(basename $(dirname $file))
    local f=$(basename $file)
    local reservation="$LANDER_QUEUE_RESERVED_FS/$lander/$f-$_lander_queue_salt"
    /bin/rm -rf $reservation
    return 0
}
