#!/bin/bash 
#PBS -l walltime=2:00:00
#PBS -k n
#PBS -o /dev/null
#PBS -e /dev/null
#PBS -m a

# $Id: user_worker.sh 15605 2013-01-24 20:01:35Z yuri $

# This script is meant to be scheduled to run on the cluster
#+by user_dispatcher.sh.  This is just a wrapper around the
#+environment $USERSCRIPT which is called to execute with the filename
#+as an argument.  If $USERSCRIPT exits with 0, the input filename will
#+be removed; otherwise, it will stay in $INPROCESSDIR directory.

THISSCRIPT=$0

LANDER_CONFIG=~plander/lander.conf
#read config
if [ -e $LANDER_CONFIG ]; then
    . $LANDER_CONFIG
else
    echo "ERROR: Config file $CONFIG does not exist" >&2
    exit 1
fi

#EXPORTED VARIABLE from user_dispatcher.sh: WORKER_ARGS
worker_args=${WORKER_ARGS:-$@}

usage() {
    cat >&2 <<EOF
    USAGE: $THISSCRIPT [OPTIONS] <indir> <procdir> <script>"
	where:
               <OPTIONS>        can be:

                 -j <jobs>      Jobs per core * 100 (ignore cores if unset)
                 -r <regex>     File regex for incoming dir
                 -s             Skip out of order files (only useful with -n 1)
                 -u             Ignore (don't try to match sequence numbers)
                 -w X           Require that the modtime at the head of the queue be 
                                at least X seconds in the past (defaults to 0)
               <indir>          Incoming directory (usually incoming_(host|flow|full|raw))
               <procdir>        In-process directory (usually inprocess_(host|flow|full|raw))
               <script>         User script to run

    Normally, this script is invoked indirectly via user_dispatcher.sh.
EOF
}

jobspercore=$LANDER_USERJOB_JOBSPERCORE #default value

minfileage=$LANDER_USERJOB_MINFILEAGE

set -- $worker_args
while getopts "hj:r:suw:" opt; do
    case $opt in
	( j ) jobspercore=$OPTARG; ;;
	( r ) FILE_REGEX=$OPTARG; ;;
	( s ) SKIPOOO="yes"; ;;
	( u ) UNORDERED="yes"; ;;
	( w ) minfileage=$OPTARG; ;;
	( h|* ) usage; exit 0; ;;
    esac
done

shift $((OPTIND-1));

INDIR=$1
INPROCESSDIR=$2
USERSCRIPT=$3
file_regex=${FILE_REGEX:-'[0-9]{8}-[0-9]{6}-[0-9]+[^/]*$'}

if [ -z "$INDIR" -o -z "$INPROCESSDIR" -o -z "$USERSCRIPT" ]; then
    echo "-i, -p, and -u are required" >&2
    exit 1
fi

USERSCRIPT_base=${USERSCRIPT##*/}
LOGFILE=${LANDER_USERDIR}/${USERSCRIPT_base}-w.log    #write to this log in user's home

exec >> $LOGFILE 2>&1

WORKERID="$(hostname -s)-$$"

#check if -j is given
if [ -n "$jobspercore" ] && [ -z "$LANDER_WORKER_FORKED" ]; then
    CPUINFO=/proc/cpuinfo
    if [ -e $CPUINFO ]; then
	ncores=$(grep ^process $CPUINFO | wc -l)
	njobs=$((ncores*jobspercore/100))
	njobs=$(( (njobs < 1)?1:njobs ))
	export LANDER_WORKER_FORKED="yes" #do not fork again
	forks=0
	for (( jobs=0; jobs<njobs; ++jobs )); do
	    $THISSCRIPT $worker_args &
	    let '++forks'
	done
	echo "*** forked $forks times" >&2
	wait
	echo "all forks completed" >&2
	exit 0
    else
	echo "$CPUINFO file doesn't exist, ignoring -j option given" >&2;
    fi
fi
LS=/bin/ls
RM=/bin/rm
MV=/bin/mv

#preemptive exit via signalling
must_exit=""
trap "must_exit=y" SIGHUP

check_load() {
    #make sure the load is OK
    MAXTRIES=10
    for (( try=0; try < MAXTRIES; ++try )); do
        #first check if the load is too high
	$set_LOAD
	[ -z "$LOAD" ] && break;
	echo "Current load: $LOAD"
	diff=$(echo $LANDER_USERJOB_MAXLOAD - $LOAD | bc -l)
	if [ -z "${diff/-*/}" ]; then
	    echo "Warning: load limit ($LANDER_USERJOB_MAXLOAD) is exceeded, sleeping"
	    #randomize sleeping time a little
	    sleep $((30 + RANDOM % 90))
	    now=$(date +%s)
	    [ $now -lt $TIMEDONE ] || exit 0
	    [ -z $must_exit ] || exit 1
	else
	    break
	fi
    done
    [ $try -ge $MAXTRIES ] && exit 0
}


now=$(date +%s)
# We quit after LANDER_USERJOB_RUN_SEC
let 'TIMEDONE = now + LANDER_USERJOB_RUN_SEC'

while true; do
    # Check if we still have files to process, limit with "head" to avoid huge lists
    files=$(find $INDIR/ -regextype posix-extended -regex ".*\/$file_regex" -print 2>/dev/null)
    if [ -z "$files" ]; then
	#sleep some, then retry, if still nothing THEN exit
	sleep 5s
	files=$(find $INDIR/ -regextype posix-extended -regex ".*\/$file_regex" -print 2>/dev/null)
	[ -z "$files" ] && exit 0
    fi
    fnkey=$(echo "/x$files" 2>/dev/null | head -1 | tr '/' ' ' | wc -w)
    sortedfiles=$(echo "$files" | sort -t '/' -k $fnkey)
    for FULLFILE in $sortedfiles; do
	#check for exitting signals
	[ -z $must_exit ] || exit 1

	#make sure file is there
	[ -e "$FULLFILE" ] || continue

	#isolate basename() and basename(dirname())
	FILE=${FULLFILE##*/}
	DIRNAME=${FULLFILE%/*}
	DIR=${DIRNAME##*/}

	#check if next sequence number has been written
	if [ -z "$UNORDERED" ]; then
	    tmp=${FILE%-*}
	    seqnum=${tmp##*-}

	    #check if the file name makes sense
	    if [ -n "${seqnum//[0-9]}" ]; then
		echo "ERROR: File name of $FILE does not seem to have a sequence number" >&2
		echo "ERROR:    You can use -u option to disable sequence number matching" >&2
		exit 1
	    fi

	    if [ -n "$SKIPOOO" ]; then
	        #may need to skip this file if it's out of order
		SEQFILE="$DIRNAME/.seqnum"
		SEQFILELOCK=$SEQFILE.lock
		TIMEOUT=60
		lockfile -1 -r $((TIMEOUT+1)) -l $TIMEOUT  $SEQFILELOCK
		read seq xxx < $SEQFILE
		if [ "$xxx" != "xxx" ]; then
		    echo "$SEQFILE is either corrupt or empty, resetting..."
		    seq=0
		fi
		if (( 10$seq <=  10#$seqnum )); then
		    if (( 10$seq + 1 != 10#$seqnum )); then
			echo "$DIRNAME: SKIPPING ($seq..$seqnum)"
		    fi
		    echo "$seqnum xxx" >$SEQFILE
		    $RM -f $SEQFILELOCK
		else
		    echo "$DIRNAME: $seqnum is too late (processed $seq), skipping"
		    #XXX rm this one?
		    $RM -f $SEQFILELOCK
		    continue
		fi
	    fi

	    nseqlen=-${#seqnum}
	    nextseqnum="000000000000$(( 10#$seqnum+1 ))"
	    nextseqnum=${nextseqnum:$nseqlen}

	    if ! ls $DIRNAME/*-$nextseqnum-* >&/dev/null; then
		modtime=$(stat -c '%Y' $FULLFILE)
		if (( delta=(minfileage - (now - modtime)), delta>0 )); then
                    #file at the head is too young
		    echo "File at the head $FULLFILE is too young: $now - $modtime < $minfileage"
		    sleep $delta
		    break 2 #re-generate file list
		else
		    echo "WARN: working on $FULLFILE, while $nextseqnum hasn't been delivered"
		fi
	    fi
	fi #-z "$UNORDERED"
	
	#make name unique
	inpfile="$INPROCESSDIR/$DIR/$FILE" 
	inpfile_tmp="$inpfile.$WORKERID"
	$MV "$FULLFILE" "$inpfile_tmp" 2>/dev/null

	# XXX sometimes mv exit status is zero, but move in fact
	#+XXX fails (probably because of NFS attribute caching), 
	#+XXX so we have to check for existence of the target.
	[ -e "$inpfile_tmp" ] || continue

	# move was successful, we work on this file
	#+by invoking users' script; we cleanup if the script
	#+exit status is true
	
	#rename to discard jobid (no longer needed)
	$MV $inpfile_tmp $inpfile

	if $USERSCRIPT $inpfile; then
	    lander_set_DATE
	    echo "$DATE $USERSCRIPT on $inpfile - success"
	    $RM -f $inpfile
	else
	    exit_status=$?
	    lander_set_DATE
	    echo "$DATE $USERSCRIPT on $inpfile - failure"
	    $MV $inpfile "$INPROCESSDIR/$DIR/$FILE"
	    # Furthermore, if the user script was killed, we propagate the signal.
	    #+Killing is indicated by exit status >127
	    if [ $exit_status -gt 127 ]; then
		echo "$DATE exitting on signal $((exit_status-127))"
		exit $exit_status
	    fi
	fi

	now=$(date +%s)
	[ $now -ge $TIMEDONE ] && exit 0

	check_load

    done #for FULLFILE in $sortedfiles
done #while true
